# ICARUS PMT and trigger ####################################################### {#ICARUSPMTTrigger}

| Document name:  | ICARUSPMTtrigger                                |
| --------------- | ----------------------------------------------- |
| Type:           | Software overview                               |
| Author:         | Gianluca Petrillo (petrillo@slac.stanford.edu)  |
| Created on:     | April 15, 2020                                  |
| Version:        | 0.1                                             |


This document gives an overview of the PMT and trigger simulation in ICARUS.


## How to use this document ####################################################

The sections of this document can be read independently.

Note that the document was written with doxygen 1.8.15 as reference.

The format is [Markdown format], that you can convert it to something else with:
    
    pandoc -s -S --toc -o ICARUSPMTtrigger.html ICARUSPMTtrigger.md
    pandoc -s -S --toc -o ICARUSPMTtrigger.pdf ICARUSPMTtrigger.md
    
et cetera.
Note that this Markdown file is processed by `doxygen`.
This means that typing in the document complete doxygen commands (either with
`@` or with `\`) will trigger a doxygen action. For that reason, the commands
names here are printed _with a prefix `\@`_ (e.g. `\@defgroup`).
Waiting for a better solution.


## Available modules and algorithms ############################################ {#ICARUSPMTTriggerAlgorithms}

This is a partial list of the available modules and algorithms for the optical
detector and trigger simulation:

* optical detector:
    * `icarus::opdet::PMTsimulationAlg` (module: `opdet::SimPMTIcarus`):
      simulation of the digitization of PMT signals
* trigger:
    * module `icarus::trigger::DiscriminatePMTwaveforms` discriminates the
      signal from each PMT;
    * module `icarus::trigger::LVDSgates` coalesces the discriminated PMT gate
      signals into combined single trigger requests;
    * module `icarus::trigger::SlidingWindowTrigger` coalesces the single
      trigger requests into sliding window multilevel gate signals;
    * module `icarus::trigger::MajorityTriggerSimulation` writes a simulated
      LArSoft trigger object based on plain single trigger request counts;
    * module `icarus::trigger::MajorityTriggerEfficiencyPlots` saves plots of
      efficiency for triggers based on plain single trigger request counts;
    * module `icarus::trigger::SlidingWindowTriggerEfficiencyPlots` saves plots
      of efficiency for triggers based on single trigger request counts within
      specific sliding windows;
    * module `icarus::trigger::MakeTriggerSimulationTree` creates a simple ROOT
      tree with information of each event and simulated trigger
    * module `icarus::trigger::DumpTriggerGateData` prints on console the
      content of a data product with `icarus::trigger::TriggerGateData` objects.



## Glossary of terms ########################################################### {#ICARUSPMTTriggerGlossary}

board-common trigger
:   a signal generated in the V1730B board combining the single trigger requests
    with a customizable logic; this signal is used in self-triggering mode to
    schedule the storage of current PMT channel data.

gate signal
:   a digital signal as function of time; the signal can be binary (e.g. levels
    0 and 1) or multilevel (e.g. levels 0, 1, 2, ...); the time is usually
    measured in clock ticks, and in case of ICARUS optical detector and trigger
    simulation, that clock is the V1730B optical digitization clock (500 MHz).

single trigger request
:   a _gate signal_ generated by the V1730B readout board from two PMT channel
    signals. Given the two PMT signals `A` and `B`, the single trigger request
    signal is generated discriminating the two signals against their own ADC
    threshold, an combining them as AND or OR or picking only `A` or only `B`.
    In a V1730B board there are 8 single trigger request signals which are used
    to generate a board-common trigger signal (if the board has self triggering
    enabled) and are also converted into LVDS for output.

trigger primitive
:   in general, each unit of information contributing to the trigger decision;
    in ICARUS, the trigger primitives are mostly single trigger requests, which
    get combined by the trigger logic hardware to return a final trigger
    decision.



